#include <stdlib.h>
//#include <stdio.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/power.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>

#include "lookups.h"			//lookup tables

#define PACKETSIZE 90
#define INCOMING_ERROR '#'

#define SET_DTR sbi(PORTD,1)
#define CLEAR_DTR cbi(PORTD,1)

#define RADIO_ON sbi(PORTB,2)
#define RADIO_OFF cbi(PORTB,2)
#define RADIO_IS_ON (PORTB&(1<<2))
									//	\/ baud rate
#define BAUDDIV  (u16)( ((float)F_CPU/(9600*16UL)) -1 )//baud divider

#define BV(bit)			(1<<(bit))
#define cbi(reg,bit)	reg &= ~(BV(bit))
#define sbi(reg,bit)	reg |= (BV(bit))
#define	outb(addr, data)	addr = (data)
#define	inb(addr)			(addr)

typedef unsigned char u08;
typedef signed char s08;
typedef unsigned int u16;
typedef unsigned long u32;
typedef struct { u08 bytes[3]; u08 size;} varicoder_type;
extern const unsigned short varicode_lookup[256];//the lookup table for varicode - the 4 msb are the number of valid bits in the data
volatile u08 bufferfront,bufferback,buffer[256],PWM;
void set_PWM(u08 set);
//void put_char(char c, FILE * stream);
//void get_char(FILE * stream);
//FILE uart_stdout = FDEV_SETUP_STREAM(put_char,get_char, _FDEV_SETUP_RW);

static u08 cycbuf[16]; 
static u08 cycptr;   // points at last position in buffer 
static u08 encstate; // encoder state - rolling input
static u32 varibuf;  // large enough to allow for worst case varicode lengths 
static s08 varibits; // count of bits in buffer

uint8_t mcusr_mirror __attribute__ ((section (".noinit")));
 
void get_mcusr(void) __attribute__((naked)) __attribute__((section(".init3")));//stops poss wdt loop
void get_mcusr(void)
{
	mcusr_mirror = MCUSR;
	MCUSR = 0;
	wdt_disable();
}	

// MFSK_Init() - initialize data structures

void MFSK_Init (void)
{
	int i;
	for (i = 0; i < 16; i++)
		cycbuf[i] = 0;
	cycptr = 0;
	encstate = 0;
	varibuf = 0;
	varibits = 0;
}

// passed a nibble (4 bits)
// calls DAC to transmit the bits

void GrayEncode(u08 arg)		// gray encodes a nibble
{
	PWM=(pgm_read_byte(&GreyTab[arg >> 4]))|0x80;// high nibble
	while(PWM);					//wait for the PWM to be free
	PWM=(pgm_read_byte(&GreyTab[arg & 0x0F]))|0x80;// low nibble
	while(PWM);					//wait for the PWM to be free
}

// Interleave() - Bit Interveiver 
// implments the equivelent of 10 sequential 4 bit diagonal interleavers
// equivelent is an offset of 10 rather than 1 (i.e. offset of 41 bits rather than 5 bits)
//
// Example Input bits: "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789ABCDE..."
//
// |A|E I M Q U Y 2 6 A E I M Q U Y 2 6 A E I M Q U Y 2 6 A E I M Q U Y 2 6 
//  B F J N R V Z 3 7 B|F|J N R V Z 3 7 B F J N R V Z 3 7 B F J N R V Z 3 7 
//  C G K O S W 0 4 8 C G K O S W 0 4 8 C G|K|O S W 0 4 8 C G K O S W 0 4 8 
//  D H L P T X 1 5 9 D H L P T X 1 5 9 D H L P T X 1 5 9 D H L|P|T X 1 5 9 
//
//  0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3 3 3 3 3 3 
//  1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 
//
// Output bits: AFKPEJOT INSXMRW1 
//
// uses a 16 byte (128bit) cyclic buffer
// feed a byte in - get a interleved byte out.
// needs 16 bytes into this function to flush everything

void Interleave(u08 input) 
{ 
	u08 output; // no need to initiallize as fully written
        cycbuf[cycptr & 0x0F] = input; // cycptr now points at the buffer base.
        output = (cycbuf[cycptr & 0x0F] & 0x88) | (cycbuf[(cycptr + 5) & 0x0F] & 0x44) | 
                (cycbuf[(cycptr + 10) & 0x0F] & 0x22) | (cycbuf[(cycptr + 15) & 0x0F] & 0x11);
        --cycptr; // cycptr now points at where the next byte will go on next call
	GrayEncode(output); // bit significance preserved.
}


// FEC_Encode() - encode a byte into 2 bytes using r=1/2, k=7 convolution
// needs 9 bytes into this function to flush interleaver

void FEC_Encode(u08 input)
{
	u16 output; // no need to initialize as fully written
	int i;
	for(i = 7; i >= 0; i--)
	{ // for each bit in the input
	     encstate = (encstate << 1) | ((input >> i) & 1); // shift the bits in - msb 1st
	     output = (output << 2) | (u16) pgm_read_byte(&PolyTab[encstate & 0x7F]); // 7 bit table
	}
	Interleave((u08)(output >> 8));	// hi byte
	Interleave((u08)(output & 0x00FF)); 	// lo byte
}



void VariEncode(u08 ascii) //varicodes input character and calls FEC when it has enough Input
{
	u08 length;
	u16 value;
	value = pgm_read_word(&VariTab[ascii]); //the upper 4 msb tell us how many valid bits there are for this character
	length = (u08)(value >> 12);	// extract length
	value &= 0x0FFF;		// strip length to just get bits 
	varibuf <<= length;	// make room at bottom
	varibuf |= (u32)value;	// add new varicode character in at the bottom
	varibits += length;	// count total bits in buffer
	while(varibits >= 8)
	{ // while there are complete bytes transmit them
		varibits -= 8;
		FEC_Encode((u08)(varibuf >> varibits));
	}
	// 0 to 7 bits remain
}

// VariFlush() - flush varicode, FEC and interleave buffers with zeros
// results in idle being sent (continious lowest tone)
//
// transmission starts with at least 8 symbols of the lowest code
// transmission ends with at least 4 symbols of the lowest code

void VariFlush(void) 
{
	int i;
	if (varibits > 0)
	{ // some varicode to flush
	     varibits -= 8;
	     varibuf <<= varibits; // shift in 0s
	     FEC_Encode((u08)(varibuf & 0x00FF)); // send remaining bits
	     varibits = 0;
	}
	for (i = 0; i < 9; i++)
	{
	     FEC_Encode(0); // send enough to flush all buffering
	}
}

/*void put_char(char c, FILE * stream)
{
	while ( !( UCSR0A & (1<<UDRE0)) );
	UDR0=c;
}

void get_char(FILE * stream)
{;}*/

ISR(USART_RX_vect)
{
	if(!(UCSR0A & (BV(FE0)|BV(UPE0))))//a framing or parity error causes us to drop the byte
		buffer[bufferfront]=UDR0;	
	else
		buffer[bufferfront]=INCOMING_ERROR;//tells us theres corruption on the data
	bufferfront++;
	if(bufferfront==bufferback)
		bufferback++;				//our buffer was filled
	if((u08)((s08)bufferback-(s08)bufferfront)<PACKETSIZE)//flow control
		CLEAR_DTR;
}


ISR(TIMER1_OVF_vect)			//fires at the same time as end of pwm pulse
{
	static u16 counter;
	counter++;
	if(counter==3148)			//1/62.5 seconds at 12MHz
	{
		counter=0;
		OCR1AL=PWM<<2;			//set the pwm to pwm*4
		PWM=0;					//so we know its been set
	}
}

int main(void)
{
	char c;
	u08 n=0;
	power_adc_disable();		//turn off unneeded modules
	power_spi_disable();
	power_timer0_disable();
	power_timer2_disable();
	power_twi_disable();
	DDRB=0b00000110;			//radio enable and pwm pins
	DDRD=0b00000010;			//use TX as a DTR pin
	PORTD=0b000001100;			//counter noise on interrupt pins
	outb(UCSR0B, BV(RXCIE0)|BV(RXEN0));//mega xx8 registers - enable tx,rx with interrupts
	outb(UCSR0C, BV(UPM01)|BV(UPM00)|BV(UCSZ01)|BV(UCSZ00));//setup 8O1
	outb(UBRR0L, BAUDDIV);
	outb(UBRR0H, BAUDDIV>>8);	//end of UART setup
	//Timer 1 setup
	sbi(TCCR1B,CS10);			//clock at F_CPU
	TCNT1 = 0;					//reset timer1  
 	// set fast PWM mode with ICR top-count
	cbi(TCCR1A,WGM10);
	sbi(TCCR1A,WGM11);
	sbi(TCCR1B,WGM12);
	sbi(TCCR1B,WGM13);
 	// set top count value to give correct period PWM
	ICR1 = 60;
 	// clear output compare value A
	OCR1A = 0;
	// turn on channel A (OC1A) PWM output
	// set OC1A as non-inverted PWM
	sbi(TCCR1A,COM1A1);
	cbi(TCCR1A,COM1A0);
	//enable overflow interrupts
	outb(TIMSK1, BV(TOIE1));	//end of pwm setup
	wdt_enable(WDTO_1S);
	//stdout=&uart_stdout;
	//enable the radio
	SET_DTR;
	RADIO_ON;
	__asm__ __volatile__ ("sei" ::);//enable global interrupts
	MFSK_Init();				//init ready for the greeting
	do
	{
		c=pgm_read_byte(&greeting[n]);
		VariEncode(c);
		n++;
		wdt_reset();
	}while(c);					//send out the greeting
	for(;;)						//main loop
	{
		if(bufferback==bufferfront)
		{
			if(RADIO_IS_ON)
			{
				//run the shutdown procedure
				VariFlush();
				RADIO_OFF;
			}
			//turn off timer1 interrupt
			outb(TIMSK1, 0x00);
			wdt_reset();					//sets register to zero
			wdt_enable(WDTO_8S);
			set_sleep_mode(SLEEP_MODE_IDLE);//hibernate until we get some data or the timer times out
			sleep_mode();
			wdt_reset();
			wdt_enable(WDTO_1S);
			outb(TIMSK1, BV(TOIE1));		//go back to normal operation
		}
		else
		{
			if(!RADIO_IS_ON)
			{
				//run the startup procedure
				RADIO_ON;
				MFSK_Init();
			}
			VariEncode(buffer[bufferback]);//send something from the buffer
			bufferback++;		//move to next buffer location		
		}
		wdt_reset();
		if((u08)((s08)bufferback-(s08)bufferfront)>PACKETSIZE)//flow control - set if we have space
			SET_DTR;
	}
}
